#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <allegro.h>

#include "audio.h"
#include "tipi.h"
#include "cpu.h"
#include "keyboard.h"
#include "debug.h"
#include "vdc.h"
#include "vpp.h"
#include "timefunc.h"
#include "voice.h"
#include "vmachine.h"

static Byte X_Latch,Y_Latch;
static Byte Line_Count;
static int Fps=FPS_NTSC;

static Byte SnapedLines[MAXLINES+2*MAXSNAP][256][2];

int EvblClk=EVBLCLK_NTSC;

struct Resource App_Data;

int Frame=0;
Byte DbStick1,DbStick2;
int Int_Clk;
int Master_Clk;
int H_Clk;
unsigned long Clk_Counter;
int Last_Line;
int Key2vCnt=0;
int MState;
int CcolFlag=0;

int pEndIrq=0;
int EnahIrq=1;
int UseForEn=0;
long RegionOff=0xffff;
int MxSnap=2;
int SprOff=0;
int TweakedAudio=0;

Byte Rom1[5120];
Byte Rom2[5120];
Byte Rom3[5120];
Byte Rom4[5120];

Byte IntRam[64];
Byte ExtRam[256];
Byte ExtRom[1024];
Byte VDCWrite[256];
Byte ColorVector[MAXLINES];
Byte AudioVector[MAXLINES];
Byte *Rom,*oRom;

int Key2[128];

static unsigned int Key_Map[6][8]= {
	{KEY_0,KEY_1,KEY_2,KEY_3,KEY_4,KEY_5,KEY_6,KEY_7},
	{KEY_8,KEY_9,0,0,KEY_SPACE,KEY_SLASH,KEY_L,KEY_P},
	{KEY_PLUS_PAD,KEY_W,KEY_E,KEY_R,KEY_T,KEY_U,KEY_I,KEY_O},
	{KEY_Q,KEY_S,KEY_D,KEY_F,KEY_G,KEY_H,KEY_J,KEY_K},
	{KEY_A,KEY_Z,KEY_X,KEY_C,KEY_V,KEY_B,KEY_M,KEY_STOP},
	{KEY_MINUS,KEY_ASTERISK,KEY_SLASH_PAD,KEY_EQUALS,KEY_Y,KEY_N,KEY_DEL,KEY_ENTER}
};

static void Do_Kluges(void);
static void SetVideoMode(int T);


void Run(void)
{
	while(!Key_Done)
	{
		if (Key_Debug)
		{
			App_Data.Debug=1;
			
			Set_TextMode();
			Mute_Audio();
			Mute_Voice();
			
			Debug();
			
			GrMode();
			
			App_Data.Debug=0;
			
			Init_Keyboard();
			Init_Sound_Stream();
		}
		
		Cpu_Exec();
	}
	
	Close_Audio();
	Close_Voice();
	Close_Display();
}

void Handle_Vbl(void)
{
	if (!App_Data.Debug)
	{
		Handle_Key();
		UpDate_Audio();
		UpDate_Voice();
	}
	
	if (!CcolFlag)
	{
		Clear_Collision();
		CcolFlag=1;
	}
	
	Draw_Region();
	Ext_IRQ();
	MState=1;
}

void Handle_Evbl(void)
{
	static long last=0;
	static int rest_cnt=0;
	int i;
	
	i=(15 * App_Data.Speed / 100);
	rest_cnt=(rest_cnt + 1) % (i < 5 ? 5 : i);
	Last_Line=0;
	Master_Clk-=EvblClk;
	Frame++;
	CcolFlag=0;
	
	if (!App_Data.Debug)
		Finish_Display();
	
	for (i=0; i < MAXLINES; i++)
	{
		ColorVector[i]=(VDCWrite[0xA3] & 0x7f) | (P1 & 0x80);
		AudioVector[i]=VDCWrite[0xAA];
	}
	
	if (Key2vCnt++ > 10)
	{
		Key2vCnt=0;
		
		for (i=0; i < 128; i++)
			Key2[i]=0;
		
		DbStick1=DbStick2=0;
	}
	
	if (App_Data.Limit)
	{
		long d,f;
		
		d=(TICKSPERSEC*100) / (App_Data.Speed * Fps);
		f=((d - GetTimeTicks() + last) * 1000) / TICKSPERSEC;
		
		while (GetTimeTicks() - last < d);
			
		last=GetTimeTicks();
	}
	
	MState=0;
}

void Init_System(void)
{
	int i,j,k;
	
	Last_Line=0;
	DbStick1=0x00;
	DbStick2=0x00;
	MState=0;
	Master_Clk=0;
	H_Clk=0;
	Line_Count=0;
	ITimer=0;
	Clk_Counter=0;
	
	for(i=0; i < 256; i++)
	{
		VDCWrite[i]=0;
		ExtRam[i]=0;
	}
	
	for(i=0; i < 64; i++)
		IntRam[i]=0;
	
	for (i=0; i < MAXLINES; i++)
		AudioVector[i]=ColorVector[i]=0;
	
	CcolFlag=0;
	
	for (i=0; i < MAXLINES + 2 * MAXSNAP; i++)
		for (j=0; j < 256; j++)
			for (k=0; k < 2; k++)
				SnapedLines[i][j][k]=0;
			
	if (App_Data.Stick[0] == 2 || App_Data.Stick[1] == 2)
	{
		i=install_joystick(JOY_TYPE_AUTODETECT);
		
		if (i || (num_joysticks < 1))
		{
			fprintf(stderr,"Error: no joystick detected\n");
			exit(EXIT_FAILURE);			
		}
	}
	
	for (i=0; i < 128; i++)
		Key2[i]=0;
	
	Key2vCnt=0;
	
	if (App_Data.Euro)
		SetVideoMode(1);
	else
		SetVideoMode(0);
	
	Do_Kluges();	
	Init_Vpp();
}

Byte Read_T1(void)
{
	if ((H_Clk > 16) || (Master_Clk > VBLCLK))
		return 1;
	else
		return 0;
}

void Write_P1(Byte D)
{
	if ((D & 0x80) != (P1 & 0x80))
	{
		int i,l;
		
		l=SnapLine((int)((float)Master_Clk / 22.0+0.5),VDCWrite[0xA3],1);
		
		for (i=l; i < MAXLINES; i++)
			ColorVector[i]=(VDCWrite[0xA3] & 0x7f) | (D & 0x80);
	}
	
	P1=D;
	
	if (App_Data.Bank == 2)
	{
		if (P1 & 0x01)
			Rom=Rom2;
		else
			Rom=Rom1;
	}
	
	if (App_Data.Bank == 3)
	{
		switch(P1 & 0x03)
		{
			case 0:
				Rom=Rom1;
				break;
			
			case 1:
				Rom=Rom2;
				break;
			
			case 2:
				Rom=Rom3;
				break;
			
			case 3:
				Rom=Rom4;
				break;
		}
	}
}

Byte Read_P2(void)
{
	int i,si,so,km;
	
	if (NeedsPool)
		poll_keyboard();
	
	if (!(P1 & 0x04))
	{
		si=(P2 & 7);
		so=0xff;
		
		if (si < 6)
		{
			for (i=0; i < 8; i++)
			{
				km=Key_Map[si][i];
				
				if ((key[km] && ((!JoyKeysTab[km]) || (key_shifts & KB_CAPSLOCK_FLAG))) || (Key2[km]))
					so=i ^ 0x07;
			}			
		}
		
		if (so != 0xff)
		{
			P2&=0x0F;
			P2|=(so << 5);
		}
		else
			P2|=0xF0;
	}
	else
		P2|=0xF0;
	
	return(P2);
}

Byte Ext_Read(Indirizzo Adr)
{
	Byte d,si,m;
	int i;
	
	if (!(P1 & 0x08) && !(P1 & 0x40))
	{
		switch(Adr)
		{
			case 0xA1:
				d=VDCWrite[0xA0] & 0x02;
			
				if (Master_Clk > VBLCLK)
					d|=0x08;
				
				if (H_Clk < (LINECNT-7))
					d|=0x01;
				
				if (Sound_IRQ)
					d|=0x04;
				
				return d;
				
			case 0xA2:
				si=VDCWrite[0xA2];
				m=0x01;
				d=0;
				
				for(i=0; i < 8; i++)
				{
					if (si & m)
					{
						if (ColTab[1] & m)
							d|=(ColTab[1] & (m ^ 0xFF));
						
						if (ColTab[2] & m)
							d|=(ColTab[2] & (m ^ 0xFF));
						
						if (ColTab[4] & m)
							d|=(ColTab[4] & (m ^ 0xFF));
						
						if (ColTab[8] & m)
							d|=(ColTab[8] & (m ^ 0xFF));
						
						if (ColTab[0x10] & m)
							d|=(ColTab[0x10] & (m ^ 0xFF));
						
						if (ColTab[0x20] & m)
							d|=(ColTab[0x20] & (m ^ 0xFF));
						
						if (ColTab[0x80] & m)
							d|=(ColTab[0x80] & (m ^ 0xFF));
					}
					
					m<<=1;
				}
				
				return d;
				
			case 0xA5:
				if (!(VDCWrite[0xA0] & 0x02))
					return X_Latch;
				else
				{
					X_Latch=H_Clk * 12;
					return X_Latch;
				}
				
			case 0xA4:
				if (!(VDCWrite[0xA0] & 0x02))
					return Y_Latch;
				else
				{
					Y_Latch=Master_Clk / 22;
					
					if (Y_Latch > 241)
						Y_Latch=0xFF;
					
					return Y_Latch;
				}
				
			default:
				return VDCWrite[Adr];
		}
	}
	else
		if (!(P1 & 0x10))
			return ExtRam[Adr & 0xFF];
		else
			if (!(P1 & 0x20))
				return Vpp_Read(Adr);
			else
				if (App_Data.ExRom && (P1 & 0x02))
					return ExtRom[(P2 << 8) | (Adr & 0xFF)];
				
	return 0;
}

Byte In_Bus(void)
{
	Byte si=0,d=0,mode=0,jn=0;
	
	if ((P1 & 0x08) && (P1 & 0x10))
	{
		if (!(P1 & 0x04))
			si=(P2 & 7);
		
		d=0xFF;
		
		if (si == 1)
		{
			mode=App_Data.Stick[0];
			jn=0;
		}
		else
		{
			mode=App_Data.Stick[1];
			jn=1;
		}
		
		switch(mode)
		{
			case 1:
				d=KeyJoy(jn);
				break;
			
			case 2:
				poll_joystick();
				
				if (joy_up)
					d&=0xFE;
				
				if (joy_right)
					d&=0xFD;
				
				if (joy_down)
					d&=0xFB;
				
				if (joy_left)
					d&=0xF7;
				
				if (joy_b1)
					d&=0xEF;
				
				break;
		}
		
		if (si == 1)
		{
			if (DbStick1)
				d=DbStick1;
		}
		else
			if (DbStick2)
				d=DbStick2;
	}
	
	return d;
}

void Ext_Write(Byte Dat,Indirizzo Adr)
{
	int i;
	
	if (!(P1 & 0x08))
	{
		if (Adr == 0xA0)
		{
			if ((VDCWrite[0xA0] & 0x02) && !(Dat & 0x02))
			{
				Y_Latch=Master_Clk / 22;
				X_Latch=H_Clk * 12;
				
				if (Y_Latch > 241)
					Y_Latch=0xFF;
			}
			
			if ((Master_Clk <= VBLCLK) && (VDCWrite[0xA0] != Dat))
			{
				if (!CcolFlag)
				{
					Clear_Collision();
					
					CcolFlag=1;
				}
				
				Draw_Region();
			}
		}
		else
			if (Adr == 0xA2)
			{
				Clear_Collision();
				
				CcolFlag=1;
			}
			else
				if (Adr == 0xA3)
				{
					int l;
					
					l=SnapLine((int)((float)Master_Clk/22.0+0.5), Dat,1);
					
					for (i=l; i < MAXLINES; i++)
						ColorVector[i]=(Dat & 0x7f) | (P1 & 0x80);
				}
				else
					if (Adr == 0xAA)
						for (i=Master_Clk / 22; i < MAXLINES; i++)
							AudioVector[i]=Dat;
						
		VDCWrite[Adr]=Dat;
	}
	else
		if (!(P1 & 0x10) && !(P1 & 0x40))
		{
			Adr&=0xFF;
			
			if (Adr < 0x80)
				ExtRam[Adr]=Dat;
			else
			{
				if (!(Dat & 0x20))
					Reset_Voice();
				else
				{
					if (Adr == 0xE4)
						Set_Voice_Bank(0);
					else
						if ((Adr >= 0xE8) && (Adr <= 0xEF))
							Set_Voice_Bank(Adr - 0xE7);
						else
							if (((Adr >= 0x80) && (Adr <= 0xDF)) || ((Adr >= 0xF0) && (Adr <= 0xFF)))
								Trigger_Voice(Adr);
				}
			}
		}
		else
			if (!(P1 & 0x20))
				Vpp_Write(Dat,Adr);
}

static void Do_Kluges(void)
{
	switch (App_Data.Crc)
	{
		case 0xA7344D1F:
			pEndIrq=1;
			RegionOff=0;
			
			SetVideoMode(1);
			break;
		
		case 0xFB83171E:
			pEndIrq=1;
			EnahIrq=0;
			RegionOff=1;
			MxSnap=3;
		
			SetVideoMode(0);
			break;
		
		case 0x881CEAE4:
			pEndIrq=1;
			MxSnap=3;
			EvblClk=6100;
		
			SetVideoMode(1);
			break;		
		
		case 0x9E42E766:
			UseForEn=1;
			MxSnap=0;
			break;
		
		case 0x1C750349:
			UseForEn=1;
			MxSnap=0;
			break;
		
		case 0x202F2749:
			UseForEn=1;
			RegionOff=1;
		
			SetVideoMode(0);
			break;				
				
		case 0x5216771A:
			RegionOff=1;
		
			TweakedAudio=1;
			break;

		case 0x0C2E4811:
			RegionOff=12;
			break;
		
		case 0x67069924:
			RegionOff=11;
			break;
		
		case 0x44D1A8A5:
			RegionOff=11;
			break;
		
		case 0xB936BD78:
			RegionOff=12;
			break;
		
		case 0xDC30AD3D:
			RegionOff=10;
			break;
		
		case 0x7810BAD5:
			RegionOff=10;
			break;		
		
		case 0xD0BC4EE6:
			RegionOff=12;
			MxSnap=3;
			EvblClk=7642;
		
			SetVideoMode(1);
		
			TweakedAudio=1;
			break;
		
		case 0x825976A9:
			RegionOff=0;
			EvblClk=7642;
		
			SetVideoMode(1);
			break;
		
		case 0xF390BFEC:
			RegionOff=0;
			EvblClk=7642;
		
			SetVideoMode(1);
			break;
		
		case 0x3BFEF56B:
			RegionOff=1;
			MxSnap=6;
			SetVideoMode(1);
			break;
		
		case 0x9BFC3E01:
			RegionOff=10;
			MxSnap=0;
		
			SetVideoMode(0);
			break;
		
		case 0x6CEBAB74:
			RegionOff=12;
			break;
		
		case 0xE7B26A56:
			RegionOff=12;
			break;
		
		case 0xA57E1724:
			MxSnap=12;
			break;
		
		case 0xFD179F6D:
			MxSnap=3;
			break;
		
		case 0x9C9DDDF9:
			MxSnap=3;
			break;
		
		case 0x95936B07:
			MxSnap=3;
			break;
		
		case 0x39E31BF0:
			SetVideoMode(1);
			break;
		
		case 0x3351FEDA:
			SetVideoMode(1);
			break;
		
		case 0x40AE062D:
			SetVideoMode(1);
			break;
		
		case 0xD158EEBA:
			SetVideoMode(1);
			break;
		
		case 0x26B0FF5B:
			SetVideoMode(1);
			break;
		
		case 0xDF36683F:
			SetVideoMode(1);
			break;
		
		case 0xAF307559:
			SetVideoMode(1);
			break;
		
		case 0x9585D511:
			SetVideoMode(1);
			break;
		
		case 0x58FA6766:
			SetVideoMode(1);
			break;
		
		case 0x39989464:
			SetVideoMode(1);
			break;
		
		case 0xB47F3E0B:
			SetVideoMode(1);
			break;

		case 0x68560DC7:
			SetVideoMode(1);
			break;
		
		case 0x239DF97D:
			SetVideoMode(0);
			break;

		case 0x2DCB77F0:
		case 0xF6882734:
			EvblClk=9000;
		
			SetVideoMode(1);
			break;

		case 0xAFB23F89:
			TweakedAudio=1;
			break;
		
		case 0xD3B09FEC:
			SprOff=1;
			break;
	}
}

int SnapLine(int Pos,Byte Reg,int T)
{
	register int i;
	
	for (i=0; i < MxSnap; i++)
	{
		if (SnapedLines[Pos + MAXSNAP - i][Reg][T])
			return Pos - i;
		
		if (SnapedLines[Pos + MAXSNAP + i][Reg][T])
			return Pos + i;
	}
	
	SnapedLines[Pos + MAXSNAP][Reg][T]=1;
	return Pos;
}

static void SetVideoMode(int T)
{
	if (T)
	{
		EvblClk=EVBLCLK_PAL;
		Fps=FPS_PAL;
	}
	else
	{
		EvblClk=EVBLCLK_NTSC;
		Fps=FPS_NTSC;
	}
}
