#include <stdio.h>
#include <allegro.h>

#include "vmachine.h"
#include "cpu.h"
#include "voice.h"

static SAMPLE *Voices[9][128];
static int Voice_Bank=0;
static int Voice_Num=-1;
static int Voice_Addr=0;
static int Voice_Ok=0;
static int Voice_St=0;
static unsigned long Clk_Voice_Start=0;


void Load_Voice_Samples(void)
{
	int bank, sam, i, ld=0;
	char name[40];
	SAMPLE *sp=NULL;
	
	printf("Loading voice samples...  ");
	fflush(stdout);
	
	for (i=0; i < 9; i++)
		for (sam=0; sam < 128; sam++)
		{
			if (i)
				bank=0xE8 + i - 1;
			else
				bank=0xE4;
			
			sprintf(name,"voice/%02x%02x.wav",bank,sam + 0x80);
			
			Voices[i][sam]=load_sample(name);
			
			if (!Voices[i][sam])
			{
				sprintf(name,"voice/%02X%02X.WAV",bank,sam + 0x80);
				Voices[i][sam]=load_sample(name);
			}
			
			if (Voices[i][sam])
			{
				ld++;
				
				if (!sp)
					sp=Voices[i][sam];
			}
		}
		
	printf("%d samples loaded\n",ld);
		
	if (ld > 0)
	{
		Voice_Num=allocate_voice(sp);
		
		if (Voice_Num != -1)
			Voice_Ok=1;
		else
		{
			printf("  ERROR: could not allocate sound card voice\n");
			Voice_Ok=0;
		}
	}
}

void UpDate_Voice(void)
{
	if (!Voice_Ok)
		return;
	
	if (Voice_St == 2)
	{
		if (voice_get_position(Voice_Num) < 0)
		{
			if ((Voice_Bank >= 0) && (Voice_Bank<9) && (Voice_Addr >= 0x80) && (Voice_Addr <= 0xff))
			{
				if (Voices[Voice_Bank][Voice_Addr - 0x80])
				{
					reallocate_voice(Voice_Num,Voices[Voice_Bank][Voice_Addr - 0x80]);
					voice_set_volume(Voice_Num,(255 * App_Data.VVolume) / 100);
					voice_start(Voice_Num);
					
					Clk_Voice_Start=Clk_Counter;
					Voice_St=1;
				}
				else
					Voice_St=0;
			}
		}
	}
		else
			if (Voice_St == 1)
				if ((voice_get_position(Voice_Num) < 0) || (Clk_Counter - Clk_Voice_Start > 20))
					Voice_St=0;
}

void Trigger_Voice(int Addr)
{
	if (Voice_Ok)
	{
		if (Voice_St)
			UpDate_Voice();
		
		if ((Voice_St == 0) && (Voice_Bank >= 0) && 
			(Voice_Bank < 9) && (Addr >= 0x80) && (Addr <= 0xff))
		{
			Voice_Addr=Addr;
			Voice_St=2;
			
			UpDate_Voice();
		}
	}
}

void Set_Voice_Bank(int Bank)
{
	if (!Voice_Ok)
		return;
	
	if ((Bank >= 0) && (Bank <= 8))
		Voice_Bank=Bank;
}

int Get_Voice_Status(void)
{
	if (Voice_Ok)
	{
		UpDate_Voice();
		
		if (Voice_St)
			return 1;
	}
	
	return 0;
}

void Reset_Voice(void)
{
	if (Voice_Ok)
	{
		voice_stop(Voice_Num);
		
		Voice_Bank=0;
		Voice_Addr=0;
		Voice_St=0;
	}
}

void Mute_Voice(void)
{
	if (Voice_Ok)
		voice_stop(Voice_Num);
}

void Close_Voice(void)
{
	Reset_Voice();
	
	Voice_Ok=0;
}
